/************************************************************************/
/*									*/
/*  Administration of EditOperations.					*/
/*									*/
/************************************************************************/

#   include	"tedConfig.h"

#   include	<stdlib.h>
#   include	<string.h>
#   include	<stdio.h>

#   include	<appDebugon.h>

#   include	"docBuf.h"
#   include	"docEdit.h"

void docInitEditPosition(	EditPosition *	ep )
    {
    ep->epParagraphNumber= -1;
    ep->epStroff= -1;
    }

void docInitEditRange(	EditRange *	er )
    {
    er->erParagraphCount= 0;
    er->erBottomLevel= DOClevOUT;

    docInitEditPosition( &(er->erStartPosition) );
    docInitEditPosition( &(er->erEndPosition) );
    }

void docInitEditOperation(	EditOperation *	eo )
    {
    docInitSelectionScope( &(eo->eoSelectionScope) );

    docInitRectangle( &(eo->eoChangedRectangle) );
    eo->eoOldBackY1= 0;

    eo->eoIBarSelectionOld= 0;
    eo->eoMultiParagraphSelectionOld= 0;

    eo->eoParaAdjustParagraphNumber= -1;
    eo->eoParaAdjustFromLine= 0;
    eo->eoParaAdjustStroffShift= 0;
    eo->eoParaAdjustStroffUpto= 0;

    docInitEditRange( &(eo->eoReformatRange) );

    eo->eoNotesDeleted= 0;
    eo->eoNotesAdded= 0;
    eo->eoParagraphsDeleted= 0;
    eo->eoParagraphsInserted= 0;
    eo->eoFieldUpdate= FIELDdoNOTHING;

    eo->eoBd= (BufferDocument *)0;
    eo->eoVoidadd= (void *)0;
    eo->eoCloseObject= (DOC_CLOSE_OBJECT)0;

    return;
    }

/************************************************************************/
/*									*/
/*  Include a position in the range in the document that is to be	*/
/*  reformatted.							*/
/*									*/
/************************************************************************/

void docIncludePositionInReformat(	EditOperation *		eo,
					const BufferItem *	paraBi,
					int			stroff )
    {
    int			paraNr;
    EditRange *		er= &(eo->eoReformatRange);

    if  ( paraBi->biLevel != DOClevPARA )
	{ LDEB(paraBi->biLevel); return;	}

    eo->eoParaAdjustParagraphNumber= -1;
    eo->eoParaAdjustFromLine= 0;
    eo->eoParaAdjustStroffShift= 0;
    eo->eoParaAdjustStroffUpto= 0;

    paraNr= docNumberOfParagraph( paraBi );

    if  ( er->erStartPosition.epParagraphNumber < 0		||
	  paraNr < er->erStartPosition.epParagraphNumber	)
	{
	er->erStartPosition.epParagraphNumber= paraNr;
	er->erStartPosition.epStroff= stroff;
	}

    if  ( paraNr == er->erStartPosition.epParagraphNumber	&&
	  stroff < er->erStartPosition.epStroff			)
	{
	er->erStartPosition.epStroff= stroff;
	}

    if  ( er->erEndPosition.epParagraphNumber < 0		||
	  paraNr > er->erEndPosition.epParagraphNumber		)
	{
	er->erEndPosition.epParagraphNumber= paraNr;
	er->erEndPosition.epStroff= stroff;
	}

    if  ( paraNr == er->erEndPosition.epParagraphNumber	&&
	  stroff > er->erEndPosition.epStroff			)
	{
	er->erEndPosition.epStroff= stroff;
	}

    return;
    }

void docEditIncludeItemInReformatRange(	EditOperation *		eo,
					BufferItem *		bi )
    {
    EditRange *		er= &(eo->eoReformatRange);
    DocumentPosition	dp;

    eo->eoParaAdjustParagraphNumber= -1;
    eo->eoParaAdjustFromLine= 0;
    eo->eoParaAdjustStroffShift= 0;
    eo->eoParaAdjustStroffUpto= 0;

    if  ( er->erBottomLevel == DOClevOUT	||
	  er->erBottomLevel > bi->biLevel	)
	{ er->erBottomLevel= bi->biLevel; }

    if  ( docFirstPosition( &dp, bi ) )
	{ LDEB(1);	}
    else{ docIncludePositionInReformat( eo, dp.dpBi, dp.dpStroff );	}

    if  ( docLastPosition( &dp, bi ) )
	{ LDEB(1);	}
    else{ docIncludePositionInReformat( eo, dp.dpBi, dp.dpStroff );	}

    return;
    }

void docSetParagraphAdjust(	EditOperation *		eo,
				BufferItem *		paraBi,
				int			fromLine,
				int			stroffShift,
				int			stroffUpto )
    {
    EditRange *		er= &(eo->eoReformatRange);
    int			paraNr;

    if  ( paraBi->biLevel != DOClevPARA )
	{ LDEB(paraBi->biLevel); return;	}

    if  ( er->erBottomLevel == DOClevOUT )
	{ er->erBottomLevel= paraBi->biLevel; }

    if  ( eo->eoParaAdjustParagraphNumber >= 0 )
	{ docEditIncludeItemInReformatRange( eo, paraBi ); return; }

    paraNr= docNumberOfParagraph( paraBi );

    if  ( er->erStartPosition.epParagraphNumber == paraNr	&&
	  er->erEndPosition.epParagraphNumber == paraNr		)
	{
	eo->eoParaAdjustParagraphNumber= paraNr;
	eo->eoParaAdjustFromLine= fromLine;
	eo->eoParaAdjustStroffShift= stroffShift;
	eo->eoParaAdjustStroffUpto= stroffUpto;
	}
    else{
	docEditIncludeItemInReformatRange( eo, paraBi );
	}

    return;
    }

void docExtendParagraphAdjust(	EditOperation *		eo,
				BufferItem *		paraBi,
				int			stroffShift )
    {
    int			paraNr;

    if  ( paraBi->biLevel != DOClevPARA )
	{ LDEB(paraBi->biLevel); return;	}

    paraNr= docNumberOfParagraph( paraBi );

    if  ( eo->eoParaAdjustParagraphNumber != paraNr )
	{ LLDEB(eo->eoParaAdjustParagraphNumber,paraNr); return;	}

    eo->eoParaAdjustStroffShift += stroffShift;
    eo->eoParaAdjustStroffUpto += stroffShift;

    return;
    }

void docEditShiftReformatRangeParaNr(	EditOperation *		eo,
					int			from,
					int			by )
    {
    EditRange *		er= &(eo->eoReformatRange);

    if  ( from < by )
	{ LLDEB(from,by);	}

    if  ( eo->eoParaAdjustParagraphNumber >= 0	&&
	  eo->eoParaAdjustParagraphNumber >= from	)
	{ eo->eoParaAdjustParagraphNumber += by;	}

    if  ( er->erStartPosition.epParagraphNumber >= 0	&&
	  er->erStartPosition.epParagraphNumber >= from	)
	{ er->erStartPosition.epParagraphNumber += by;	}

    if  ( er->erEndPosition.epParagraphNumber >= 0	&&
	  er->erEndPosition.epParagraphNumber >= from	)
	{ er->erEndPosition.epParagraphNumber += by;	}

    er->erParagraphCount += by;

    if  ( er->erStartPosition.epParagraphNumber >= 0			&&
	  er->erStartPosition.epParagraphNumber > er->erParagraphCount	)
	{ er->erStartPosition.epParagraphNumber=  er->erParagraphCount; }

    if  ( er->erEndPosition.epParagraphNumber >= 0			&&
	  er->erEndPosition.epParagraphNumber > er->erParagraphCount	)
	{ er->erEndPosition.epParagraphNumber=  er->erParagraphCount; }

    if  ( er->erStartPosition.epParagraphNumber == 0	&&
	  er->erParagraphCount > 0			)
	{ er->erStartPosition.epParagraphNumber= 1;	}

    if  ( er->erEndPosition.epParagraphNumber == 0	&&
	  er->erParagraphCount > 0			)
	{ er->erEndPosition.epParagraphNumber= 1;	}

    return;
    }
